'use strict';

const through = require('through2');
const path = require('path');
const {makeArray, isFunction, isString, lopped, insertFunctions} = require('./lib/util');

const exportedRequire = insertFunctions(
	_require.toString().replace(/_require/g, 'require'),
	'// insert-functions',
	isFunction, isString, lopped, resolve
);

/**
 * Resolve a path to a given root.
 *
 * @param {path} to			Path to resolve to.
 * @param {path} from		Path to resolve from.
 * @returns {string}		Resolved path.
 */
function resolve(to, from) {
	const path = lopped(to).concat(from.split('/'));
	const resolved = [];
	let back = 0;
	for (let n=path.length-1; n>=0; n--) {
		if (path[n] !== '.') {
			if (path[n] === '..') {
				back++;
			} else if (back > 0) {
				back--;
			} else {
				resolved.unshift(path[n]);
			}
		}
	}

	return resolved.join('/');
}

/**
 * Call the given functions to insert text before and after file contents.
 *
 * @param {VinylFile} file		The vinyl-file we are dealing with.
 * @param {Function} pre		Pre function to get Buffer or text for preppending.
 * @param {Function} post		Post function to get Buffer or text for appending.
 */
function wrapVinyl(file, pre, post) {
	if (file.isStream()) {
		let prepended = false;
		file.contents = file.contents.pipe(through2(
			{decodeStrings:false, encoding:'utf8'},
			function(chunk, encoding, callback) {
				if (!prepended) {
					this.push(pre(true));
					prepended = true;
				}
				callback();
			}, function(done) {
				this.push(post(true));
				done();
			}
		));
	} else if (file.isBuffer()) {
		file.contents = new Buffer(pre()+file.contents.toString('utf8')+post());
	}
}

/**
 * Create a pre/post function that returns the given text either as a Buffer or just text depending on
 * boolean parameter.
 *
 * @param {string} prePost		Text for pre/post function
 * @returns {Function}			Pre/Post function.
 */
function prePost(prePost) {
	/**
	 * @param {booleam} [buffer=true]		Whether to serve a Buffer or just plain text.
	 */
	return (buffer=true)=>(buffer?new Buffer(prePost):prePost);
}

/**
 * Pipe function for wrapping a module.
 *
 * @param {Object} options		The current plugin options.
 * @param {VinylFile} file		The vinyl-file we are dealing with.
 * @param {string} encoding		The file encoding.
 * @param {function} callback	Pipe callback to fire.
 */
function pluginRequires(options, file, encoding, callback) {
	if (file.isStream() || file.isBuffer()) {
		const moduleId = './' + path.relative(file.cwd, file.path);
		wrapVinyl(file, prePost('require(function(require, module){'), prePost(`}, '${moduleId}');`));
	}
	return callback(null, file);
}

/**
 * Pipe function for wrapping an entire module (ie. the root module).
 *
 * @param {Object} options		The current plugin options.
 * @param {VinylFile} file		The vinyl-file we are dealing with.
 * @param {string} encoding		The file encoding.
 * @param {function} callback	Pipe callback to fire.
 */
function pluginModule(options, file, encoding, callback) {
	if (file.isStream() || file.isBuffer()) {
		const requires = makeArray(options.main).map(main=>`require("${main}");`).join('');
		wrapVinyl(
			file,
			prePost(`(function(){const _commonjsBrowserWrapModules = new Map();${exportedRequire}`),
			prePost(`${requires}})()`)
		);
	}
	return callback(null, file);
}

function _require(moduleFunction, moduleId) {
	const _moduleId = moduleIdFix(moduleId);
	const module = {};

	// insert-functions

	/**
	 * Add .js to end of module-id and ./ to beginning (where they are not present).
	 *
	 * @param {string} moduleId		Module id to fix
	 * @returns {string}			Fixed module id.
	 */
	function moduleIdFix(moduleId) {
		const _moduleId = ((!isFunction(moduleFunction))?moduleFunction:moduleId).replace(/\.js$/, '');
		return (((_moduleId.charAt(0) !== '/') && (_moduleId.charAt(0) !== '.')) ? './' + _moduleId : _moduleId);
	}

	/**
	 * Get a require function scoped to module.
	 *
	 * @returns {Function}		Scoped require.
	 */
	function getLocalRequire() {
		return (localModuleId, ...params)=>_require(
			isString(_moduleId)?resolve(_moduleId, localModuleId):localModuleId,
			...params
		);
	}

	if (isFunction(moduleFunction)) return _commonjsBrowserWrapModules.set(_moduleId, moduleFunction);
	if (!_commonjsBrowserWrapModules.has(_moduleId)) throw new SyntaxError(`Cannot find module with id: ${moduleId}`);
	_commonjsBrowserWrapModules.get(_moduleId)(getLocalRequire(_moduleId), module);
	return module.exports;
}

module.exports = options=>{
	return through.obj(function (...params){
		if (!options || !options.type || (options.type === 'requireWrap')) {
			pluginRequires.bind(this)(options, ...params);
		} else if (options && (options.type === 'moduleWrap')) {
			pluginModule.bind(this)(options, ...params);
		}
	});
};